import os
from datetime import datetime


class ResourceException(Exception):
	pass


class Resource(object):


	def resolve_attr(self, attr_name, attr_val, default=None, raises=True):

		if not hasattr(self, attr_name) or getattr(self, attr_name) is None:
			setattr(self, attr_name, attr_val)

		if getattr(self, attr_name) is None:
			setattr(self, attr_name, default)

		if getattr(self, attr_name) is None and raises:
			raise ResourceException('No value found for %s.' % attr_name)

		return getattr(self, attr_name)



		
	def get_ready(self, lot, as_pilot, name, clobber):

		# resolve propogation of the run properties
		self.resolve_attr('lot', lot)
		self.resolve_attr('as_pilot', as_pilot, False)
		self.resolve_attr('name', name)
		self.resolve_attr('clobber', clobber)

		# make sure that the lot is string-like
		if not isinstance(self.lot, basestring):
			raise ValueError(
				'`lot` must be an instance of basestr, like str or unicode.'
			)

		self._is_ready = True


	def is_ready(self):
		try:
			return self._is_ready
		except AttributeError:
			return False


	def exists(self):
		raise NotImplementedError(
			'Resources need to implement an `exists` method.'
		)


class FileResource(Resource):

	def __init__(self, path, fname):
		self.path = path
		self.fname = fname


	def get_path(self):

		# ensure the resource is ready
		if not self.is_ready():
			raise ResourceException('Resource is not ready.')

		fname = (
			self.lot + '_' 
			+ ('pilot' if self.as_pilot else '') 
			+ self.fname
		)
		return os.path.join(self.path, fname)


	def exists(self):
		return os.path.isfile(self.get_path())

	def open(self, flags):

		# there's some things to check if we're writing to a *new* file
		if 'w' in flags:

			# if writing, check whether the file exists
			if os.path.isfile(self.get_path()):
				raise IOError(
					'FileResource: by default, I refuse to overwrite files. '
					+ self.get_path()
				)

			# check if the directory exists, if not make it
			if not os.path.exists(self.path):
				os.makedirs(self.path)

			# ensure that the directory is not an existing file
			if os.path.isfile(self.path):
				raise IOError(
					'FileResource: the given path corresponds to an existing '
					'*file*.'
				)

		# hands over a plain file handle
		return open(self.get_path(), flags)


class MarkerResource(FileResource):
	def mark(self):

		# if the marker path already exists, perfect
		if os.path.isdir(self.get_path()):
			pass

		# if the marker path points to a file, not a dir, fail
		elif os.path.exists(self.get_path()):
			raise IOError(
				('MarkerResource: I could not mark task `%s` as done because '
				'there is a file in the place of the directory I am supposed '
				'to write to') % self.name
			)

		# if the path doesn't exist yet, make it
		else:
			os.mkdirs(self.path)

		self.open('a').write('%s\n' % str(datetime.now()))


class IncrementalFileResource(object):
	'''
		This is like a file resource, but we don't rely on the existence of
		the file to determine whether the resource is satisfied.  We use a
		separate "marker" file, that is created when the resource is marked
		as done (by calling mark_as_done()).  This file contains a timestamp
		of when the resource was marked done according to local machine's time
	'''

	def __init__(self, path, fname, part_name):
		self.path = path
		self.fname = fname
		self.part_name = part_name

		
	def exists(self):
		return os.path.isfile(get_marker_path)


	def get_marker_path(self):
		return self.get_path() + '.' + self.part_name + '.marker'


	def mark_as_done(self):
		# note that this will create a marker file with the current time
		# stamp.  In general, it should never be called if there is an existing
		# file already by the same name, however, if that does happen somehow
		# it will append to the file.
		marker_fh = open(self.get_marker_path(), 'a')
		marker_fh.write('%s\n' % str(datetime.now()))




class FolderResource(FileResource):
	'''
		Creates (if necessary) a folder that is namepsaced to the lot, 
		and allows reading and writing files there.  File names are not
		namespaced (because the folder is).
	'''

	def get_fname(self):
		return os.path.join(self.get_path(), fname)

	def open(self, fname, mode):

		# check if the folder exists yet, make it if not 
		if os.path.isdir(self.get_path()):
			pass

		# if a file (rather than folder) exists, it's an error
		elif os.exists(self.get_path()):
			raise IOError(
				'FolderResource: the given path corresponds to an existing '
				'*file*.  I cannot create a directory here: %s' 
				% self.get_path()
			)

		# if not, make the folder
		else:
			os.mkdir(self.get_path())

		# check if the specific file exists (as a folder or file)
		if os.path.isdir(self.get_fname()):
			raise IOError(
				'FolderResource: a file or folder exists there. '
				'I can\'t make a file: %s' % self.get_fname()
			)

		# if we're opening in write mode, don't overwrite an existing file
		if 'w' in mode and os.isfile(self.get_fname()):
			raise IOError(
				'FolderResource: a file already exists there. '
				'I don not overwrite by default: %s' % self.get_fname()
			)

		# if all is good, open and yield the file resource
		return open(self.get_fname(), mode)


	#TODO: defined exists
	def exists(self):
		pass




import shutil
import unittest
from unittest import TestCase
from run import Runner, RunnerException
from task import Task, TaskException, MarkedTask, SimpleTask
from resource import Resource, FileResource
import os


TEST_DIR = 'temporary_testing_files'


def touch(fname):
	open(fname, 'a').close()


class TestRunner(TestCase):

	def test_null_runner(self):

		# if you don't define a lot, an exception is raised
		runner = Runner()
		with self.assertRaises(RunnerException):
			runner.run()

		# A null runner can be run, once you specify the lot
		class MyRunner(Runner):
			lot = 'my_lot'

		runner = MyRunner()
		runner.run()


	def test_run_until(self):

		class MyTask(Task):
			outputs = None
			was_run = False		# Track whether the task was run
			def exists(self):	# ensures the task will not be skipped
				return False
			def run(self):
				self.was_run = True

		task0 = MyTask()
		task1 = MyTask()
		task2 = MyTask()
		task3 = MyTask()

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {
				'task0':task0,
				'task1':task1,
				'task2':task2,
				'task3':task3
			}
			layout = {
				'END': 'task3',
				'task3': 'task2',
				'task2': ['task1','task0']
			}

		my_runner = MyRunner()
		my_runner.run(until='task2')

		self.assertTrue(task0.was_run)
		self.assertTrue(task1.was_run)
		self.assertTrue(task2.was_run)
		self.assertFalse(task3.was_run)


	def test_run_until_classdef(self):

		class MyTask(Task):
			outputs = None
			was_run = False		# Track whether the task was run
			def exists(self):	# ensures the task will not be skipped
				return False
			def run(self):
				self.was_run = True

		task0 = MyTask()
		task1 = MyTask()
		task2 = MyTask()
		task3 = MyTask()

		class MyRunner(Runner):
			lot = 'my_lot'
			until = 'task2'
			tasks = {
				'task0':task0,
				'task1':task1,
				'task2':task2,
				'task3':task3
			}
			layout = {
				'END': 'task3',
				'task3': 'task2',
				'task2': ['task1','task0']
			}

		my_runner = MyRunner()
		my_runner.run()

		self.assertTrue(task0.was_run)
		self.assertTrue(task1.was_run)
		self.assertTrue(task2.was_run)
		self.assertFalse(task3.was_run)


	def test_run_until_end(self):

		class MyTask(Task):
			outputs = None
			was_run = False		# Track whether the task was run
			def exists(self):	# ensures the task will not be skipped
				return False
			def run(self):
				self.was_run = True

		task0 = MyTask()
		task1 = MyTask()
		task2 = MyTask()
		task3 = MyTask()

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {
				'task0':task0,
				'task1':task1,
				'task2':task2,
				'task3':task3
			}
			layout = {
				'END': 'task3',
				'task3': 'task2',
				'task2': ['task1','task0']
			}

		my_runner = MyRunner()
		my_runner.run()

		self.assertTrue(task0.was_run)
		self.assertTrue(task1.was_run)
		self.assertTrue(task2.was_run)
		self.assertTrue(task3.was_run)


class TestMarkerTask(TestCase):
	def setUp(self):
		os.mkdir(TEST_DIR)

	def tearDown(self):
		shutil.rmtree(TEST_DIR)

	def test_mark_marks(self):

		
		class MyTask(MarkedTask):
			marker_path = TEST_DIR
			outputs = None
			num_runs = 0
			def run(self):
				self.num_runs += 1

		task1 = MyTask()
		task_name = 'task1'
		lot_name = 'my_lot'

		class MyRunner(Runner):
			lot = lot_name
			tasks = {
				task_name: task1
			}
			layout = {
				'END': task_name
			}

		my_runner = MyRunner()

		# run the runner
		my_runner.run()
		# check that the task was run
		self.assertTrue(task1.num_runs == 1)

		# run again
		my_runner.run()
		# this task should have been skipped due to existence of marker file
		self.assertTrue(task1.num_runs == 1)

		# remove the marker file
		os.remove(
			os.path.join(TEST_DIR, '%s_%s.marker' % (lot_name, task_name))
		)

		# run again
		my_runner.run()
		# the task should run again
		self.assertTrue(task1.num_runs == 2)


	def test_marker_as_mixin(self):
		'''
			this test is similar to the last, but it tests a situation where
			a pre-existing task is augmented using the mixin approach.
			rather than MyTask inheriting directly from MarkedTask,
			MyTask is defined on top of Task, and then an augmented version
			is made afterward, MyMarkedTask, by using MarkedTask as a mixin.

			This tests all the conditions in test_marker_marks, but it also
			tests that if MyTask overrides a method in Task, augmenting 
			with MarkedTask as a mixin preserves that.

			Of course, the `exists()` function is not preserved -- changing 
			that is the whole reason for using the MarkedTask as a mixin.
		'''
		
		class MyTask(Task):
			ready_override = False
			after_override = False
			def get_ready(self, lot, as_pilot, name, clobber):
				super(MyTask, self).get_ready(
					lot, as_pilot, name, clobber
				)
				self.ready_override = True

			def _after(self):
				super(MyTask, self)._after()
				self.after_override = True


		class MyMarkedTask(MarkedTask, MyTask):
			marker_path = TEST_DIR
			outputs = None
			num_runs = 0
			def run(self):
				self.num_runs += 1

		task1 = MyMarkedTask()
		task_name = 'task1'
		lot_name = 'my_lot'

		class MyRunner(Runner):
			lot = lot_name
			tasks = {
				task_name: task1
			}
			layout = {
				'END': task_name
			}

		my_runner = MyRunner()

		# run the runner
		my_runner.run()
		# check that the task was run
		self.assertTrue(task1.num_runs == 1)

		# check that the version of get_ready from MyTask was run
		self.assertTrue(task1.ready_override)
		# check that the version of after from MyTask was run
		self.assertTrue(task1.after_override)

		# run again
		my_runner.run()
		# this task should have been skipped due to existence of marker file
		self.assertTrue(task1.num_runs == 1)

		# remove the marker file
		os.remove(
			os.path.join(TEST_DIR, '%s_%s.marker' % (lot_name, task_name))
		)

		# run again
		my_runner.run()
		# the task should run again
		self.assertTrue(task1.num_runs == 2)


class TestSimpleTask(TestCase):

	TEST_DIR = 'linguini_markers'
	def setUp(self):
		os.mkdir(self.TEST_DIR)

	def tearDown(self):
		shutil.rmtree(self.TEST_DIR)

	def test_simple_task(self):
		'''
			Simple task is based on marked task, and provides the simplest
			interface for turning some piece of functionality into a linguini
			task.  One simply defines the run function, that's all
		'''

		class MyTask(SimpleTask):
			num_runs = 0
			def run(self):
				self.num_runs += 1

		task1 = MyTask()
		task_name = 'task1'
		lot_name = 'my_lot'

		class MyRunner(Runner):
			lot = lot_name
			tasks = {
				task_name: task1
			}
			layout = {
				'END': task_name
			}

		my_runner = MyRunner()

		# run the runner
		my_runner.run()
		# check that the task was run
		self.assertTrue(task1.num_runs == 1)

		# run again
		my_runner.run()
		# this task should have been skipped due to existence of marker file
		self.assertTrue(task1.num_runs == 1)

		# remove the marker file
		os.remove(
			os.path.join(self.TEST_DIR, '%s_%s.marker' % (lot_name, task_name))
		)

		# run again
		my_runner.run()
		# the task should run again
		self.assertTrue(task1.num_runs == 2)



class TestCompoundRunner(TestCase):

	def setUp(self):
		os.mkdir(TEST_DIR)

	def tearDown(self):
		shutil.rmtree(TEST_DIR)

	def test_run_compound(self):
		'''
			Tests that a compound runner does run the entire thing
		'''
		
		class MyTask(Task):

			def _outputs(self):
				fname = 'test%d.txt' % self.parameters['file_num']
				return FileResource(TEST_DIR, fname)

			was_run = False		# Track whether the task was run
			def exists(self):	# ensures the task will not be skipped
				return False

			def run(self):
				fh = self.outputs.open('w')
				fh.write('yo')
				fh.close()
				self.was_run = True

		task0 = MyTask(file_num=0)
		task1 = MyTask(file_num=1)
		task2 = MyTask(file_num=2)
		task3 = MyTask(file_num=3)
		task4 = MyTask(file_num=4)
		task5 = MyTask(file_num=5)
		task6 = MyTask(file_num=6)
		task7 = MyTask(file_num=7)

		class MyRunner(Runner):
			outputs = None
			lot = 'a'
			tasks = {
				'task0':task0,
				'task1':task1,
				'task2':task2,
				'task3':task3
			}
			layout = {
				'END': 'task3',
				'task3': 'task2',
				'task2': ['task1','task0']
			}
			def exists(self):
				return False

		class MoRunner(Runner):
			outputs = None
			lot = 'b'
			tasks = {
				'task0':task4,
				'task1':task5,
				'task2':task6,
				'task3':task7
			}
			layout = {
				'END': 'task3',
				'task3': 'task2',
				'task2': ['task1','task0']
			}
			def exists(self):
				return False

		class MasterRunner(Runner):
			lot = '1'
			tasks = {
				'run0': MyRunner(),
				'run1': MoRunner()
			}
			layout = {
				'END': 'run1',
				'run1': 'run0'
			}

		my_runner = MasterRunner()
		my_runner.run()

		self.assertTrue(task0.was_run)
		self.assertTrue(task1.was_run)
		self.assertTrue(task2.was_run)
		self.assertTrue(task3.was_run)
		self.assertTrue(task4.was_run)
		self.assertTrue(task5.was_run)
		self.assertTrue(task6.was_run)
		self.assertTrue(task7.was_run)

		expected_files = ['a_test%d.txt' % d for d in range(4)]
		expected_files.extend(['b_test%d.txt' % d for d in range(4,8)])
		self.assertItemsEqual(expected_files, os.listdir(TEST_DIR))


	def test_run_compound_partially_complete(self):
		'''
			Tests that a compound runner does runs the entire but checks
			whether the components have been run before
		'''
		
		class MyTask(Task):

			def _outputs(self):
				fname = 'test%d.txt' % self.parameters['file_num']
				return FileResource(TEST_DIR, fname)

			was_run = False		# Track whether the task was run

			def run(self):
				fh = self.outputs.open('w')
				fh.write('yo')
				fh.close()
				self.was_run = True

		task0 = MyTask(file_num=0)
		task1 = MyTask(file_num=1)
		task2 = MyTask(file_num=2)
		task3 = MyTask(file_num=3)
		task4 = MyTask(file_num=4)
		task5 = MyTask(file_num=5)
		task6 = MyTask(file_num=6)
		task7 = MyTask(file_num=7)

		class MyRunner(Runner):
			outputs = None
			lot = 'a'
			tasks = {
				'task0':task0,
				'task1':task1,
				'task2':task2,
				'task3':task3
			}
			layout = {
				'END': 'task3',
				'task3': 'task2',
				'task2': ['task1','task0']
			}
			def exists(self):
				return False

		class MoRunner(Runner):
			outputs = None
			lot = 'b'
			tasks = {
				'task0':task4,
				'task1':task5,
				'task2':task6,
				'task3':task7
			}
			layout = {
				'END': 'task3',
				'task3': 'task2',
				'task2': ['task1','task0']
			}
			def exists(self):
				return False

		my_runner = MyRunner()
		mo_runner = MoRunner()


		class MasterRunner(Runner):
			lot = '1'
			tasks = {
				'run0': my_runner,
				'run1': mo_runner
			}
			layout = {
				'END': 'run1',
				'run1': 'run0'
			}

		# make some resources so they won't be run
		touch(os.path.join(TEST_DIR, 'a_test0.txt'))
		touch(os.path.join(TEST_DIR, 'b_test6.txt'))

		my_runner = MasterRunner()
		my_runner.run()

		# these tasks should not have been run
		self.assertFalse(task0.was_run)
		self.assertFalse(task4.was_run)
		self.assertFalse(task5.was_run)
		self.assertFalse(task6.was_run)

		# these tasks should have been run
		self.assertTrue(task1.was_run)
		self.assertTrue(task2.was_run)
		self.assertTrue(task3.was_run)
		self.assertTrue(task7.was_run)

		# all the files should exist at this point
		expected_files = ['a_test%d.txt' % d for d in range(4)]
		expected_files.extend(['b_test%d.txt' % d for d in range(6,8)])

		# note, b_test4.txt and b_test5.txt never get made because
		# they aren't needed.

		self.assertItemsEqual(expected_files, os.listdir(TEST_DIR))





class TestResource(TestCase):
	

	def test_exists_not_implemented(self):
		'''
			Test that a task that doesn't define a run function raises a 
			not implemented error
		'''
		
		class MyResource(Resource):
			pass

		class MyTask(Task):
			outputs = MyResource()

			def run(self):
				print 'hello world'

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':MyTask()}
			layout = {'END':'my_task'}

		runner = MyRunner()
		with self.assertRaises(NotImplementedError):
			runner.run()


	def test_exists_implemented(self):
		'''
			Test that a task that doesn't define a run function raises a 
			not implemented error
		'''
		
		class MyResource(Resource):
			def exists(self):
				return True

		class MyTask(Task):
			outputs = MyResource()

			def run(self):
				print 'hello world'

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':MyTask()}
			layout = {'END':'my_task'}

		runner = MyRunner()
		runner.run()



class TestTask(TestCase):

	def test_no_output(self):	

		# a task has to have some outputs
		class MyTask(Task):
			def run(self):
				print 'hello world'

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':MyTask()}
			layout = {'END':'my_task'}

		runner = MyRunner()

		with self.assertRaises(TaskException):
			runner.run()


	def test_run_not_implemented(self):
		'''
			Test that a task that doesn't define a run function raises a 
			not implemented error
		'''
		
		class MyResource(Resource):
			def exists(self):
				return False

		class MyTask(Task):
			outputs = Resource()

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':MyTask()}
			layout = {'END':'my_task'}

		runner = MyRunner()

		with self.assertRaises(NotImplementedError):
			runner.run()


	def test_simple_task_gets_run(self):
		'''
			Test that a task that doesn't define a run function raises a 
			not implemented error
		'''
		
		class MyResource(Resource):
			def exists(self):
				return False

		class MyTask(Task):
			outputs = MyResource()
			was_run = False

			def run(self):
				self.was_run = True

		my_task = MyTask()

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':my_task}
			layout = {'END':'my_task'}

		runner = MyRunner()
		runner.run()

		self.assertTrue(my_task.was_run)


	def test_simple_task_not_run(self):
		'''
			Test that a task whose only resource already exists does not get 
			run.
		'''
		
		class MyResource(Resource):
			def exists(self):
				return True

		class MyTask(Task):
			outputs = MyResource()
			was_run = False

			def run(self):
				self.was_run = True

		my_task = MyTask()

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':my_task}
			layout = {'END':'my_task'}

		runner = MyRunner()
		runner.run()

		self.assertFalse(my_task.was_run)



	def test_simple_task_many_resources_dict_run(self):
		'''
			Test that a task with many resources as a list,
			gets run if at least one of them does not exist
		'''
		
		class Exists(Resource):
			def exists(self):
				return True

		class Noexists(Resource):
			def exists(self):
				return False

		class MyTask(Task):
			outputs = {
				'one': Exists(),
				'two': Noexists(),
				'three': Exists(),
				'four': Exists()
			}
			was_run = False

			def run(self):
				self.was_run = True

		my_task = MyTask()

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':my_task}
			layout = {'END':'my_task'}

		runner = MyRunner()
		runner.run()

		self.assertTrue(my_task.was_run)


	def test_simple_task_many_resources_dict_no_run(self):
		'''
			Test that a task with many resources as a list, not 
			gets run if at least one of them does not exist
		'''
		
		class Exists(Resource):
			def exists(self):
				return True

		class MyTask(Task):
			outputs = {
				'one': Exists(),
				'two': Exists(),
				'three': Exists(),
				'four': Exists()
			}
			was_run = False

			def run(self):
				self.was_run = True

		my_task = MyTask()

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':my_task}
			layout = {'END':'my_task'}

		runner = MyRunner()
		runner.run()

		self.assertFalse(my_task.was_run)


	def test_simple_task_many_resources_no_run(self):
		'''
			Test that a task with many resources as a list, not 
			gets run if at least one of them does not exist
		'''
		
		class Exists(Resource):
			def exists(self):
				return True


		class MyTask(Task):
			outputs = (Exists(), Exists(), Exists(), Exists())
			was_run = False

			def run(self):
				self.was_run = True

		my_task = MyTask()

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':my_task}
			layout = {'END':'my_task'}

		runner = MyRunner()
		runner.run()

		self.assertFalse(my_task.was_run)




	def test_simple_task_many_resources(self):
		'''
			Test that a task with many resources as a list, not 
			gets run if at least one of them does not exist
		'''
		
		class Exists(Resource):
			def exists(self):
				return True

		class Noexists(Resource):
			def exists(self):
				return False

		class MyTask(Task):
			outputs = (Exists(), Exists(), Noexists(), Exists())
			was_run = False

			def run(self):
				self.was_run = True

		my_task = MyTask()

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':my_task}
			layout = {'END':'my_task'}

		runner = MyRunner()
		runner.run()

		self.assertTrue(my_task.was_run)





if __name__ == '__main__':
	unittest.main()




from task import Task
from resource import Resource, FileResource

def as_list(item):
	if isinstance(item, dict):
		return item.values()

	if hasattr(item, '__iter__'):
		return list(item)

	return [item]


class RunnerException(Exception):
	pass


class Runner(Task):

	# The null runner has no tasks, but is valid
	tasks = {}
	layout = {
		'END':[]
	}
	until = None
	lot = None


	# setting these here makes the call to super.get_ready work, after which
	# inputs and outputs are set to the source ind sink tasks' inputs and 
	# outputs.  Any values actually set to these class variables would be 
	# ignored
	outputs = None
	inputs = None

	def _tasks(self):
		return self.tasks

	def _layout(self):
		return self.layout

	def _until(self):
		return self.until


	def check_layout(self):
		# ensure that all tasks in layout are defined in tasks
		laid_out_tasks = set([t for t in self.layout if t is not 'END'])

		missing = [
			t for t in self.layout 
			if t not in self.tasks and t is not 'END'
		]
		if len(missing)>0:
			raise RunnerException(
				'The following tasks were defined in the layout, but are '
				'not defined in the tasks: %s' % ', '.join(missing)
			)

		unused = [t for t in self.tasks if t not in self.layout]
		if len(missing)>0:
			raise RunnerException(
				'Warning, the following tasks are defined, but not included '
				'in the layout (they will never be run): %s'
				% ', '.join(unused)
			)


	def _until_names(self):
		'''
			Get the names of the last tasks to be run.  Default is the task(s) 
			marked in layout by the key 'END'; can be overridden by class 
			variable.
		'''
		return as_list(self.until or self.layout['END'])


	def _until_tasks(self):
		'''
			Get the last tasks to be run.
		'''
		return [self.tasks[t] for t in self._until_names()]


	def get_all_outputs(self):
		'''
			Collect all the outputs from the tasks that end this runner.
			These are taken to be the outputs of the runner itself.
		'''
		return reduce(
			lambda x,y: x+y.get_all_outputs(),
			self._until_tasks(),
			[]
		)


	def get_ready(
			self,
			lot=None,
			as_pilot=False,
			name='main',
			clobber=False
		):

		if self.is_ready():
			return

		# unlike other tasks, by default, runs do *not* inherit lot from parent
		# runs, unless explicitly done at object instantiation.
		# note that, at the time that get_ready is run, any lot value set 
		# during instantiation will have already been set to self.lot
		if not hasattr(self, 'lot') or self.lot is None:
			raise RunnerException('No lot was given!')

		# resolve propogation of the run properties
		self.resolve_attr('lot', lot)
		self.resolve_attr('as_pilot', as_pilot, False)
		self.resolve_attr('name', name)
		self.resolve_attr('clobber', clobber)

		# make sure that the lot is string-like
		if not isinstance(self.lot, basestring):
			raise ValueError(
				'`lot` must be an instance of basestr, like str or unicode.'
			)

		# resolve tasks
		self.tasks = self._tasks()

		# get all the tasks ready
		for task_name, task in self.tasks.iteritems():
			task.get_ready(
				lot=self.lot, as_pilot=as_pilot, name=task_name, 
				clobber=clobber
			)

		# resolve layout
		self.layout = self._layout()

		# it isn't necessary to specify a run's outputs, because they are
		# taken to be those of the END tasks
		self.outputs = self.get_all_outputs()

		# mark as ready
		self._is_ready = True


	def recursively_schedule(self, task_names):
		'''
			accepts the name of a task, or an iterable thereof, and schedules
			those which are not yet done.
		'''
		task_names = as_list(task_names)
		scheduled = []
		for task_name in reversed(task_names):

			if not self.tasks[task_name].exists():
				scheduled.append(task_name)

				# schedule dependant jobs, if any
				try:
					scheduled.extend(
						self.recursively_schedule(self.layout[task_name])
					)
				except KeyError:
					pass
				
		return scheduled


	def _run(self, lot=None, as_pilot=False, until=None, clobber=None):

		self.run(lot, as_pilot=as_pilot, until=until, clobber=clobber)


	def run(self, lot=None, as_pilot=None, until=None, clobber=False):

		self.get_ready(lot, as_pilot, 'main', clobber)

		# resolve until
		if until is not None:
			self.until = until

		elif not hasattr(self, 'until'):
			self.until = None

		elif hasattr(self, '_until'):
			self.until = self._until()

		# make sure that all the tasks are well-defined and actually get used
		self.check_layout()

		#print 'Starting runner for lot %s.' % self.lot

		# by default, run whatever is in 'END'
		until = until or self.until
		if until is None:
			until = self.layout['END']

		schedule = self.recursively_schedule(until)

		# completely run the until_tasks
		while len(schedule) > 0:
			next_task_name = schedule.pop()
			#print ' - running %s.' % next_task_name
			self.tasks[next_task_name]._run()


